@@ -1,10 +1,10 @@
 /*
   +----------------------------------------------------------------------+
   | See COPYING file for further copyright information                   |
-  +----------------------------------------------------------------------+ 
+  +----------------------------------------------------------------------+
   | Author: Oleg Grenrus <oleg.grenrus@dynamoid.com>                     |
   | See CREDITS for contributors                                         |
-  +----------------------------------------------------------------------+ 
+  +----------------------------------------------------------------------+
 */
 
 #ifdef HAVE_CONFIG_H
@@ -20,15 +20,24 @@
 #include "zend_dynamic_array.h"
 #include "zend_alloc.h"
 #include "ext/standard/info.h"
-#include "ext/session/php_session.h"
+#include "ext/standard/php_var.h"
+
+#if HAVE_PHP_SESSION
+# include "ext/session/php_session.h"
+#endif /* HAVE_PHP_SESSION */
+
 #include "ext/standard/php_incomplete_class.h"
-#ifdef HAVE_APC_SUPPORT 
+
+#if defined(HAVE_APCU_SUPPORT)
+# include "ext/apcu/apc_serializer.h"
+#elif defined(HAVE_APC_SUPPORT)
 # if USE_BUNDLED_APC
 #  include "apc_serializer.h"
 # else
 #  include "ext/apc/apc_serializer.h"
 # endif
-#endif /* HAVE_APC_SUPPORT */
+#endif /* HAVE_APCU_SUPPORT || HAVE_APC_SUPPORT */
+
 #include "php_igbinary.h"
 
 #include "igbinary.h"
@@ -45,10 +54,12 @@
 #include <stddef.h>
 #include "hash.h"
 
+#if HAVE_PHP_SESSION
 /** Session serializer function prototypes. */
 PS_SERIALIZER_FUNCS(igbinary);
+#endif /* HAVE_PHP_SESSION */
 
-#ifdef HAVE_APC_SUPPORT
+#if defined(HAVE_APC_SUPPORT) || defined(HAVE_APCU_SUPPORT)
 /** Apc serializer function prototypes */
 static int APC_SERIALIZER_NAME(igbinary) (APC_SERIALIZER_ARGS);
 static int APC_UNSERIALIZER_NAME(igbinary) (APC_UNSERIALIZER_ARGS);
@@ -123,6 +134,7 @@ struct igbinary_serialize_data {
 	struct hash_si objects;		/**< Hash of already serialized objects. */
 	int string_count;			/**< Serialized string count, used for back referencing */
 	int error;					/**< Error number. Not used. */
+	struct igbinary_memory_manager	mm; /**< Memory management functions. */
 };
 
 /** String/len pair for the igbinary_unserializer_data.
@@ -154,9 +166,14 @@ struct igbinary_unserialize_data {
 	smart_str string0_buf;			/**< Temporary buffer for strings */
 };
 /* }}} */
+/* {{{ Memory allocator wrapper prototypes */
+static inline void *igbinary_mm_wrapper_malloc(size_t size, void *context);
+static inline void *igbinary_mm_wrapper_realloc(void *ptr, size_t size, void *context);
+static inline void igbinary_mm_wrapper_free(void *ptr, void *context);
+/* }}} */
 /* {{{ Serializing functions prototypes */
-inline static int igbinary_serialize_data_init(struct igbinary_serialize_data *igsd, bool scalar TSRMLS_DC);
-inline static void igbinary_serialize_data_deinit(struct igbinary_serialize_data *igsd TSRMLS_DC);
+inline static int igbinary_serialize_data_init(struct igbinary_serialize_data *igsd, bool scalar, struct igbinary_memory_manager *memory_manager TSRMLS_DC);
+inline static void igbinary_serialize_data_deinit(struct igbinary_serialize_data *igsd, int free_buffer TSRMLS_DC);
 
 inline static int igbinary_serialize_header(struct igbinary_serialize_data *igsd TSRMLS_DC);
 
@@ -220,9 +237,30 @@ zend_function_entry igbinary_functions[] = {
 	{NULL, NULL, NULL}
 };
 /* }}} */
+
+/* {{{ igbinary dependencies */
+#if ZEND_MODULE_API_NO >= 20050922
+static const zend_module_dep igbinary_module_deps[] = {
+	ZEND_MOD_REQUIRED("standard")
+#ifdef HAVE_PHP_SESSION
+	ZEND_MOD_REQUIRED("session")
+#endif
+#if defined(HAVE_APCU_SUPPORT)
+	ZEND_MOD_OPTIONAL("apcu")
+#elif defined(HAVE_APC_SUPPORT)
+	ZEND_MOD_OPTIONAL("apc")
+#endif
+	{NULL, NULL, NULL}
+};
+#endif
+/* }}} */
+
 /* {{{ igbinary_module_entry */
 zend_module_entry igbinary_module_entry = {
-#if ZEND_MODULE_API_NO >= 20010901
+#if ZEND_MODULE_API_NO >= 20050922
+	STANDARD_MODULE_HEADER_EX, NULL,
+	igbinary_module_deps,
+#elif ZEND_MODULE_API_NO >= 20010901
 	STANDARD_MODULE_HEADER,
 #endif
 	"igbinary",
@@ -233,7 +271,7 @@ zend_module_entry igbinary_module_entry = {
 	NULL,
 	PHP_MINFO(igbinary),
 #if ZEND_MODULE_API_NO >= 20010901
-	IGBINARY_VERSION, /* Replace with version number for your extension */
+	PHP_IGBINARY_VERSION, /* Replace with version number for your extension */
 #endif
 	STANDARD_MODULE_PROPERTIES
 };
@@ -271,7 +309,7 @@ PHP_MINIT_FUNCTION(igbinary) {
 		PS_SERIALIZER_DECODE_NAME(igbinary));
 #endif
 
-#ifdef HAVE_APC_SUPPORT
+#if defined(HAVE_APC_SUPPORT) || defined(HAVE_APCU_SUPPORT)
 	apc_register_serializer("igbinary",
 		APC_SERIALIZER_NAME(igbinary),
 		APC_UNSERIALIZER_NAME(igbinary),
@@ -305,49 +343,89 @@ PHP_MINFO_FUNCTION(igbinary) {
 	(void) zend_module;
 	php_info_print_table_start();
 	php_info_print_table_row(2, "igbinary support", "enabled");
-	php_info_print_table_row(2, "igbinary version", IGBINARY_VERSION);
-#ifdef HAVE_APC_SUPPORT
+	php_info_print_table_row(2, "igbinary version", PHP_IGBINARY_VERSION);
+#if defined(HAVE_APCU_SUPPORT)
+	php_info_print_table_row(2, "igbinary APCU serializer ABI", APC_SERIALIZER_ABI);
+#elif defined(HAVE_APC_SUPPORT)
 	php_info_print_table_row(2, "igbinary APC serializer ABI", APC_SERIALIZER_ABI);
 #else
 	php_info_print_table_row(2, "igbinary APC serializer ABI", "no");
+#endif
+#if HAVE_PHP_SESSION
+	php_info_print_table_row(2, "igbinary session support", "yes");
+#else
+	php_info_print_table_row(2, "igbinary session support", "no");
 #endif
 	php_info_print_table_end();
 
 	DISPLAY_INI_ENTRIES();
 }
 /* }}} */
-/* {{{ int igbinary_serialize(uint8_t**, size_t*, zval*) */
-int igbinary_serialize(uint8_t **ret, size_t *ret_len, zval *z TSRMLS_DC) {
-	struct igbinary_serialize_data igsd;
+/* {{{ Memory allocator wrappers */
+static inline void *igbinary_mm_wrapper_malloc(size_t size, void *context)
+{
+    return emalloc(size);
+}
 
-	if (igbinary_serialize_data_init(&igsd, Z_TYPE_P(z) != IS_OBJECT && Z_TYPE_P(z) != IS_ARRAY TSRMLS_CC)) {
+static inline void *igbinary_mm_wrapper_realloc(void *ptr, size_t size, void *context)
+{
+    return erealloc(ptr, size);
+}
+
+static inline void igbinary_mm_wrapper_free(void *ptr, void *context)
+{
+    return efree(ptr);
+}
+/* }}} */
+/* {{{ int igbinary_serialize(uint8_t**, size_t*, zval*) */
+IGBINARY_API int igbinary_serialize(uint8_t **ret, size_t *ret_len, zval *z TSRMLS_DC) {
+	return igbinary_serialize_ex(ret, ret_len, z, NULL TSRMLS_CC);
+}
+/* }}} */
+/* {{{ int igbinary_serialize_ex(uint8_t**, size_t*, zval*, igbinary_memory_manager*) */
+IGBINARY_API int igbinary_serialize_ex(uint8_t **ret, size_t *ret_len, zval *z, struct igbinary_memory_manager *memory_manager TSRMLS_DC) {
+	struct igbinary_serialize_data igsd;
+	uint8_t *tmpbuf;
+
+	if (igbinary_serialize_data_init(&igsd, Z_TYPE_P(z) != IS_OBJECT && Z_TYPE_P(z) != IS_ARRAY, memory_manager TSRMLS_CC)) {
 		zend_error(E_WARNING, "igbinary_serialize: cannot init igsd");
 		return 1;
 	}
 
 	if (igbinary_serialize_header(&igsd TSRMLS_CC) != 0) {
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+		zend_error(E_WARNING, "igbinary_serialize: cannot write header");
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
 		return 1;
 	}
 
 	if (igbinary_serialize_zval(&igsd, z TSRMLS_CC) != 0) {
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
 		return 1;
 	}
 
 	/* Explicit nul termination */
-	*ret_len = igsd.buffer_size;
-	*ret = (uint8_t *) emalloc(igsd.buffer_size + 1);
-	memcpy(*ret, igsd.buffer, igsd.buffer_size);
-	*(*ret + igsd.buffer_size) = 0;
+	if (igbinary_serialize8(&igsd, 0 TSRMLS_CC) != 0) {
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
+		return 1;
+	}
 
-	igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+	/* shrink buffer to the real length, ignore errors */
+	tmpbuf = (uint8_t *) igsd.mm.realloc(igsd.buffer, igsd.buffer_size, igsd.mm.context);
+	if (tmpbuf != NULL) {
+		igsd.buffer = tmpbuf;
+	}
+
+	/* Set return values */
+	*ret_len = igsd.buffer_size - 1;
+	*ret = igsd.buffer;
+
+	igbinary_serialize_data_deinit(&igsd, 0 TSRMLS_CC);
 
 	return 0;
 }
 /* }}} */
 /* {{{ int igbinary_unserialize(const uint8_t *, size_t, zval **) */
-int igbinary_unserialize(const uint8_t *buf, size_t buf_len, zval **z TSRMLS_DC) {
+IGBINARY_API int igbinary_unserialize(const uint8_t *buf, size_t buf_len, zval **z TSRMLS_DC) {
 	struct igbinary_unserialize_data igsd;
 
 	igbinary_unserialize_data_init(&igsd TSRMLS_CC);
@@ -384,10 +462,10 @@ PHP_FUNCTION(igbinary_unserialize) {
 	}
 
 	if (string_len <= 0) {
-		RETURN_NULL();
+		RETURN_FALSE;
 	}
 
-	if (igbinary_unserialize((uint8_t *) string, string_len, &return_value TSRMLS_CC)) {
+	if (igbinary_unserialize((uint8_t *) string, string_len, &return_value TSRMLS_CC) != 0) {
 		RETURN_NULL();
 	}
 }
@@ -395,7 +473,8 @@ PHP_FUNCTION(igbinary_unserialize) {
 /* {{{ proto mixed igbinary_serialize(string value) */
 PHP_FUNCTION(igbinary_serialize) {
 	zval *z;
-	struct igbinary_serialize_data igsd;
+	uint8_t *string;
+	size_t string_len;
 
 	(void) return_value_ptr;
 	(void) this_ptr;
@@ -406,60 +485,53 @@ PHP_FUNCTION(igbinary_serialize) {
 		RETURN_NULL();
 	}
 
-	if (igbinary_serialize_data_init(&igsd, Z_TYPE_P(z) != IS_OBJECT && Z_TYPE_P(z) != IS_ARRAY TSRMLS_CC)) {
-		zend_error(E_WARNING, "igbinary_serialize: cannot init igsd");
+	if (igbinary_serialize(&string, &string_len, z TSRMLS_CC) != 0) {
 		RETURN_NULL();
 	}
 
-	if (igbinary_serialize_header(&igsd TSRMLS_CC) != 0) {
-		zend_error(E_WARNING, "igbinary_serialize: cannot write header");
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
-		RETURN_NULL();
-	}
-
-	if (igbinary_serialize_zval(&igsd, z TSRMLS_CC) != 0) {
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
-		RETURN_NULL();
-	}
-
-	RETVAL_STRINGL((char *)igsd.buffer, igsd.buffer_size, 1);
-
-	igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+	RETVAL_STRINGL((char *)string, string_len, 0);
 }
 /* }}} */
+#ifdef HAVE_PHP_SESSION
 /* {{{ Serializer encode function */
 PS_SERIALIZER_ENCODE_FUNC(igbinary)
 {
 	struct igbinary_serialize_data igsd;
-	char *s;
+	uint8_t *tmpbuf;
 
-	if (igbinary_serialize_data_init(&igsd, false TSRMLS_CC)) {
+	if (igbinary_serialize_data_init(&igsd, false, NULL TSRMLS_CC)) {
 		zend_error(E_WARNING, "igbinary_serialize: cannot init igsd");
 		return FAILURE;
 	}
 
 	if (igbinary_serialize_header(&igsd TSRMLS_CC) != 0) {
-		zend_error(E_WARNING, "igbinary_serailize: cannot write header");
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+		zend_error(E_WARNING, "igbinary_serialize: cannot write header");
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
 		return FAILURE;
 	}
 
 	if (igbinary_serialize_array(&igsd, PS(http_session_vars), false, false TSRMLS_CC) != 0) {
-		igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
 		return FAILURE;
 	}
 
-	s = estrndup((char*)igsd.buffer, igsd.buffer_size);
-	if (s == NULL) {
+	if (igbinary_serialize8(&igsd, 0 TSRMLS_CC) != 0) {
+		igbinary_serialize_data_deinit(&igsd, 1 TSRMLS_CC);
 		return FAILURE;
 	}
 
-	*newstr = s;
+	/* shrink buffer to the real length, ignore errors */
+	tmpbuf = (uint8_t *)igsd.mm.realloc(igsd.buffer, igsd.buffer_size, igsd.mm.context);
+	if (tmpbuf != NULL) {
+		igsd.buffer = tmpbuf;
+	}
+
+	*newstr = (char *)igsd.buffer;
 	if (newlen) {
-		*newlen = igsd.buffer_size;
+		*newlen = igsd.buffer_size - 1;
 	}
 
-	igbinary_serialize_data_deinit(&igsd TSRMLS_CC);
+	igbinary_serialize_data_deinit(&igsd, 0 TSRMLS_CC);
 
 	return SUCCESS;
 }
@@ -525,8 +597,9 @@ PS_SERIALIZER_DECODE_FUNC(igbinary) {
 	return SUCCESS;
 }
 /* }}} */
+#endif /* HAVE_PHP_SESSION */
 
-#ifdef HAVE_APC_SUPPORT
+#if defined(HAVE_APC_SUPPORT) || defined(HAVE_APCU_SUPPORT)
 /* {{{ apc_serialize function */
 static int APC_SERIALIZER_NAME(igbinary) ( APC_SERIALIZER_ARGS ) {
 	(void)config;
@@ -555,16 +628,25 @@ static int APC_UNSERIALIZER_NAME(igbinary) ( APC_UNSERIALIZER_ARGS ) {
 
 /* {{{ igbinary_serialize_data_init */
 /** Inits igbinary_serialize_data. */
-inline static int igbinary_serialize_data_init(struct igbinary_serialize_data *igsd, bool scalar TSRMLS_DC) {
+inline static int igbinary_serialize_data_init(struct igbinary_serialize_data *igsd, bool scalar, struct igbinary_memory_manager *memory_manager TSRMLS_DC) {
 	int r = 0;
 
+	if (memory_manager == NULL) {
+		igsd->mm.alloc = igbinary_mm_wrapper_malloc;
+		igsd->mm.realloc = igbinary_mm_wrapper_realloc;
+		igsd->mm.free = igbinary_mm_wrapper_free;
+		igsd->mm.context = NULL;
+	} else {
+		igsd->mm = *memory_manager;
+	}
+
 	igsd->buffer = NULL;
 	igsd->buffer_size = 0;
 	igsd->buffer_capacity = 32;
 	igsd->string_count = 0;
 	igsd->error = 0;
 
-	igsd->buffer = (uint8_t *) emalloc(igsd->buffer_capacity);
+	igsd->buffer = (uint8_t *) igsd->mm.alloc(igsd->buffer_capacity, igsd->mm.context);
 	if (igsd->buffer == NULL) {
 		return 1;
 	}
@@ -582,9 +664,9 @@ inline static int igbinary_serialize_data_init(struct igbinary_serialize_data *i
 /* }}} */
 /* {{{ igbinary_serialize_data_deinit */
 /** Deinits igbinary_serialize_data. */
-inline static void igbinary_serialize_data_deinit(struct igbinary_serialize_data *igsd TSRMLS_DC) {
-	if (igsd->buffer) {
-		efree(igsd->buffer);
+inline static void igbinary_serialize_data_deinit(struct igbinary_serialize_data *igsd, int free_buffer TSRMLS_DC) {
+	if (free_buffer && igsd->buffer) {
+		igsd->mm.free(igsd->buffer, igsd->mm.context);
 	}
 
 	if (!igsd->scalar) {
@@ -600,7 +682,7 @@ inline static int igbinary_serialize_header(struct igbinary_serialize_data *igsd
 }
 /* }}} */
 /* {{{ igbinary_serialize_resize */
-/** Expandes igbinary_serialize_data. */
+/** Expands igbinary_serialize_data. */
 inline static int igbinary_serialize_resize(struct igbinary_serialize_data *igsd, size_t size TSRMLS_DC) {
 	if (igsd->buffer_size + size < igsd->buffer_capacity) {
 		return 0;
@@ -610,7 +692,7 @@ inline static int igbinary_serialize_resize(struct igbinary_serialize_data *igsd
 		igsd->buffer_capacity *= 2;
 	}
 
-	igsd->buffer = (uint8_t *) erealloc(igsd->buffer, igsd->buffer_capacity);
+	igsd->buffer = (uint8_t *) igsd->mm.realloc(igsd->buffer, igsd->buffer_capacity, igsd->mm.context);
 	if (igsd->buffer == NULL)
 		return 1;
 
@@ -696,11 +778,20 @@ inline static int igbinary_serialize_long(struct igbinary_serialize_data *igsd,
 	/* -LONG_MIN is 0 otherwise. */
 	if (l == LONG_MIN) {
 #if SIZEOF_LONG == 8
-		igbinary_serialize8(igsd, (uint8_t) igbinary_type_long64n TSRMLS_CC);
-		igbinary_serialize64(igsd, (uint64_t) 0x8000000000000000 TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_long64n TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize64(igsd, (uint64_t) 0x8000000000000000 TSRMLS_CC) != 0) {
+			return 1;
+		}
 #elif SIZEOF_LONG == 4
-		igbinary_serialize8(igsd, (uint8_t) igbinary_type_long32n TSRMLS_CC);
-		igbinary_serialize32(igsd, (uint32_t) 0x80000000 TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_long32n TSRMLS_CC) != 0) {
+			return 1;
+		}
+		if (igbinary_serialize32(igsd, (uint32_t) 0x80000000 TSRMLS_CC) != 0) {
+			return 1;
+		}
 #else
 #error "Strange sizeof(long)."
 #endif
@@ -708,23 +799,46 @@ inline static int igbinary_serialize_long(struct igbinary_serialize_data *igsd,
 	}
 
 	if (k <= 0xff) {
-		igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long8p : igbinary_type_long8n) TSRMLS_CC);
-		igbinary_serialize8(igsd, (uint8_t) k TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long8p : igbinary_type_long8n) TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize8(igsd, (uint8_t) k TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else if (k <= 0xffff) {
-		igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long16p : igbinary_type_long16n) TSRMLS_CC);
-		igbinary_serialize16(igsd, (uint16_t) k TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long16p : igbinary_type_long16n) TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize16(igsd, (uint16_t) k TSRMLS_CC) != 0) {
+			return 1;
+		}
 #if SIZEOF_LONG == 8
 	} else if (k <= 0xffffffff) {
-		igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long32p : igbinary_type_long32n) TSRMLS_CC);
-		igbinary_serialize32(igsd, (uint32_t) k TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long32p : igbinary_type_long32n) TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize32(igsd, (uint32_t) k TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else {
-		igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long64p : igbinary_type_long64n) TSRMLS_CC);
-		igbinary_serialize64(igsd, (uint64_t) k TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long64p : igbinary_type_long64n) TSRMLS_CC) != 0) {
+			return 1;
+		}
+		if (igbinary_serialize64(igsd, (uint64_t) k TSRMLS_CC) != 0) {
+			return 1;
+		}
 	}
 #elif SIZEOF_LONG == 4
 	} else {
-		igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long32p : igbinary_type_long32n) TSRMLS_CC);
-		igbinary_serialize32(igsd, (uint32_t) k TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) (p ? igbinary_type_long32p : igbinary_type_long32n) TSRMLS_CC) != 0) {
+			return 1;
+		}
+		if (igbinary_serialize32(igsd, (uint32_t) k TSRMLS_CC) != 0) {
+			return 1;
+		}
 	}
 #else
 #error "Strange sizeof(long)."
@@ -741,13 +855,13 @@ inline static int igbinary_serialize_double(struct igbinary_serialize_data *igsd
 		uint64_t u;
 	} u;
 
-	igbinary_serialize8(igsd, igbinary_type_double TSRMLS_CC);
+	if (igbinary_serialize8(igsd, igbinary_type_double TSRMLS_CC) != 0) {
+		return 1;
+	}
 
 	u.d = d;
 
-	igbinary_serialize64(igsd, u.u TSRMLS_CC);
-
-	return 0;
+	return igbinary_serialize64(igsd, u.u TSRMLS_CC);
 }
 /* }}} */
 /* {{{ igbinary_serialize_string */
@@ -759,7 +873,10 @@ inline static int igbinary_serialize_string(struct igbinary_serialize_data *igsd
 	uint32_t *i = &t;
 
 	if (len == 0) {
-		igbinary_serialize8(igsd, igbinary_type_string_empty TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_string_empty TSRMLS_CC) != 0) {
+			return 1;
+		}
+
 		return 0;
 	}
 
@@ -775,14 +892,29 @@ inline static int igbinary_serialize_string(struct igbinary_serialize_data *igsd
 		}
 	} else {
 		if (*i <= 0xff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id8 TSRMLS_CC);
-			igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id8 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else if (*i <= 0xffff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id16 TSRMLS_CC);
-			igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id16 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id32 TSRMLS_CC);
-			igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_string_id32 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		}
 	}
 
@@ -793,14 +925,29 @@ inline static int igbinary_serialize_string(struct igbinary_serialize_data *igsd
 /** Serializes string data. */
 inline static int igbinary_serialize_chararray(struct igbinary_serialize_data *igsd, const char *s, size_t len TSRMLS_DC) {
 	if (len <= 0xff) {
-		igbinary_serialize8(igsd, igbinary_type_string8 TSRMLS_CC);
-		igbinary_serialize8(igsd, len TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_string8 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize8(igsd, len TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else if (len <= 0xffff) {
-		igbinary_serialize8(igsd, igbinary_type_string16 TSRMLS_CC);
-		igbinary_serialize16(igsd, len TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_string16 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize16(igsd, len TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else {
-		igbinary_serialize8(igsd, igbinary_type_string32 TSRMLS_CC);
-		igbinary_serialize32(igsd, len TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_string32 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize32(igsd, len TSRMLS_CC) != 0) {
+			return 1;
+		}
 	}
 
 	if (igbinary_serialize_resize(igsd, len TSRMLS_CC)) {
@@ -842,14 +989,29 @@ inline static int igbinary_serialize_array(struct igbinary_serialize_data *igsd,
 	}
 
 	if (n <= 0xff) {
-		igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC);
-		igbinary_serialize8(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize8(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else if (n <= 0xffff) {
-		igbinary_serialize8(igsd, igbinary_type_array16 TSRMLS_CC);
-		igbinary_serialize16(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array16 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize16(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else {
-		igbinary_serialize8(igsd, igbinary_type_array32 TSRMLS_CC);
-		igbinary_serialize32(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array32 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize32(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	}
 
 	if (n == 0) {
@@ -873,12 +1035,17 @@ inline static int igbinary_serialize_array(struct igbinary_serialize_data *igsd,
 
 		switch (key_type) {
 			case HASH_KEY_IS_LONG:
-				igbinary_serialize_long(igsd, key_index TSRMLS_CC);
+				if (igbinary_serialize_long(igsd, key_index TSRMLS_CC) != 0) {
+					return 1;
+				}
 				break;
-			case HASH_KEY_IS_STRING:
 
-				igbinary_serialize_string(igsd, key, key_len-1 TSRMLS_CC);
+			case HASH_KEY_IS_STRING:
+				if (igbinary_serialize_string(igsd, key, key_len-1 TSRMLS_CC) != 0) {
+					return 1;
+				}
 				break;
+
 			default:
 				zend_error(E_ERROR, "igbinary_serialize_array: key is not string nor array");
 				/* not reached */
@@ -896,7 +1063,6 @@ inline static int igbinary_serialize_array(struct igbinary_serialize_data *igsd,
 				return 1;
 			}
 		}
-
 	}
 
 	return 0;
@@ -930,16 +1096,31 @@ inline static int igbinary_serialize_array_ref(struct igbinary_serialize_data *i
 		enum igbinary_type type;
 		if (*i <= 0xff) {
 			type = object ? igbinary_type_objref8 : igbinary_type_ref8;
-			igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC);
-			igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else if (*i <= 0xffff) {
 			type = object ? igbinary_type_objref16 : igbinary_type_ref16;
-			igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC);
-			igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else {
 			type = object ? igbinary_type_objref32 : igbinary_type_ref32;
-			igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC);
-			igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) type TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		}
 
 		return 0;
@@ -968,14 +1149,29 @@ inline static int igbinary_serialize_array_sleep(struct igbinary_serialize_data
 
 	/* Serialize array id. */
 	if (n <= 0xff) {
-		igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC);
-		igbinary_serialize8(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize8(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else if (n <= 0xffff) {
-		igbinary_serialize8(igsd, igbinary_type_array16 TSRMLS_CC);
-		igbinary_serialize16(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array16 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize16(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	} else {
-		igbinary_serialize8(igsd, igbinary_type_array32 TSRMLS_CC);
-		igbinary_serialize32(igsd, n TSRMLS_CC);
+		if (igbinary_serialize8(igsd, igbinary_type_array32 TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		if (igbinary_serialize32(igsd, n TSRMLS_CC) != 0) {
+			return 1;
+		}
 	}
 
 	if (n == 0) {
@@ -1005,12 +1201,19 @@ inline static int igbinary_serialize_array_sleep(struct igbinary_serialize_data
 			/* we should still add element even if it's not OK,
 			 * since we already wrote the length of the array before
 			 * serialize null as key-value pair */
-			igbinary_serialize_null(igsd TSRMLS_CC);
+			if (igbinary_serialize_null(igsd TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else {
 
 			if (zend_hash_find(Z_OBJPROP_P(z), Z_STRVAL_PP(d), Z_STRLEN_PP(d) + 1, (void *) &v) == SUCCESS) {
-				igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC);
-				igbinary_serialize_zval(igsd, *v TSRMLS_CC);
+				if (igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC) != 0) {
+					return 1;
+				}
+
+				if (igbinary_serialize_zval(igsd, *v TSRMLS_CC) != 0) {
+					return 1;
+				}
 			} else if (ce) {
 				char *prot_name = NULL;
 				char *priv_name = NULL;
@@ -1021,9 +1224,15 @@ inline static int igbinary_serialize_array_sleep(struct igbinary_serialize_data
 					zend_mangle_property_name(&priv_name, &prop_name_length, ce->name, ce->name_length,
 								Z_STRVAL_PP(d), Z_STRLEN_PP(d), ce->type & ZEND_INTERNAL_CLASS);
 					if (zend_hash_find(Z_OBJPROP_P(z), priv_name, prop_name_length+1, (void *) &v) == SUCCESS) {
-						igbinary_serialize_string(igsd, priv_name, prop_name_length TSRMLS_CC);
+						if (igbinary_serialize_string(igsd, priv_name, prop_name_length TSRMLS_CC) != 0) {
+							efree(priv_name);
+							return 1;
+						}
+
 						efree(priv_name);
-						igbinary_serialize_zval(igsd, *v TSRMLS_CC);
+						if (igbinary_serialize_zval(igsd, *v TSRMLS_CC) != 0) {
+							return 1;
+						}
 						break;
 					}
 					efree(priv_name);
@@ -1032,23 +1241,39 @@ inline static int igbinary_serialize_array_sleep(struct igbinary_serialize_data
 					zend_mangle_property_name(&prot_name, &prop_name_length, "*", 1,
 								Z_STRVAL_PP(d), Z_STRLEN_PP(d), ce->type & ZEND_INTERNAL_CLASS);
 					if (zend_hash_find(Z_OBJPROP_P(z), prot_name, prop_name_length+1, (void *) &v) == SUCCESS) {
-						igbinary_serialize_string(igsd, prot_name, prop_name_length TSRMLS_CC);
+						if (igbinary_serialize_string(igsd, prot_name, prop_name_length TSRMLS_CC) != 0) {
+							efree(prot_name);
+							return 1;
+						}
+
 						efree(prot_name);
-						igbinary_serialize_zval(igsd, *v TSRMLS_CC);
+						if (igbinary_serialize_zval(igsd, *v TSRMLS_CC) != 0) {
+							return 1;
+						}
 						break;
 					}
 					efree(prot_name);
 
 					/* no win */
-					igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC);
-					igbinary_serialize_null(igsd TSRMLS_CC);
 					php_error_docref(NULL TSRMLS_CC, E_NOTICE, "\"%s\" returned as member variable from __sleep() but does not exist", Z_STRVAL_PP(d));
+					if (igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC) != 0) {
+						return 1;
+					}
+
+					if (igbinary_serialize_null(igsd TSRMLS_CC) != 0) {
+						return 1;
+					}
 				} while (0);
 
 			} else {
 				// if all else fails, just serialize the value in anyway.
-				igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC);
-				igbinary_serialize_zval(igsd, *v TSRMLS_CC);
+				if (igbinary_serialize_string(igsd, Z_STRVAL_PP(d), Z_STRLEN_PP(d) TSRMLS_CC) != 0) {
+					return 1;
+				}
+
+				if (igbinary_serialize_zval(igsd, *v TSRMLS_CC) != 0) {
+					return 1;
+				}
 			}
 		}
 	}
@@ -1067,14 +1292,29 @@ inline static int igbinary_serialize_object_name(struct igbinary_serialize_data
 		igsd->string_count += 1;
 
 		if (name_len <= 0xff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object8 TSRMLS_CC);
-			igbinary_serialize8(igsd, (uint8_t) name_len TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object8 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize8(igsd, (uint8_t) name_len TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else if (name_len <= 0xffff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object16 TSRMLS_CC);
-			igbinary_serialize16(igsd, (uint16_t) name_len TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object16 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize16(igsd, (uint16_t) name_len TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object32 TSRMLS_CC);
-			igbinary_serialize32(igsd, (uint32_t) name_len TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object32 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize32(igsd, (uint32_t) name_len TSRMLS_CC) != 0) {
+				return 1;
+			}
 		}
 
 		if (igbinary_serialize_resize(igsd, name_len TSRMLS_CC)) {
@@ -1086,14 +1326,29 @@ inline static int igbinary_serialize_object_name(struct igbinary_serialize_data
 	} else {
 		/* already serialized string */
 		if (*i <= 0xff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id8 TSRMLS_CC);
-			igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id8 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize8(igsd, (uint8_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else if (*i <= 0xffff) {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id16 TSRMLS_CC);
-			igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id16 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize16(igsd, (uint16_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		} else {
-			igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id32 TSRMLS_CC);
-			igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC);
+			if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_id32 TSRMLS_CC) != 0) {
+				return 1;
+			}
+
+			if (igbinary_serialize32(igsd, (uint32_t) *i TSRMLS_CC) != 0) {
+				return 1;
+			}
 		}
 	}
 
@@ -1125,28 +1380,65 @@ inline static int igbinary_serialize_object(struct igbinary_serialize_data *igsd
 
 	/* custom serializer */
 	if (ce && ce->serialize != NULL) {
-		/* TODO: var_hash? */
-		if(ce->serialize(z, &serialized_data, &serialized_len, (zend_serialize_data *)NULL TSRMLS_CC) == SUCCESS && !EG(exception)) {
-			igbinary_serialize_object_name(igsd, ce->name, ce->name_length TSRMLS_CC);
+		if (ce->serialize(z, &serialized_data, &serialized_len, (zend_serialize_data *)NULL TSRMLS_CC) == SUCCESS && !EG(exception)) {
+			if (igbinary_serialize_object_name(igsd, ce->name, ce->name_length TSRMLS_CC) != 0) {
+				if (serialized_data) {
+					efree(serialized_data);
+				}
+				return 1;
+			}
+
 
 			if (serialized_len <= 0xff) {
-				igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser8 TSRMLS_CC);
-				igbinary_serialize8(igsd, (uint8_t) serialized_len TSRMLS_CC);
+				if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser8 TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
+
+				if (igbinary_serialize8(igsd, (uint8_t) serialized_len TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
 			} else if (serialized_len <= 0xffff) {
-				igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser16 TSRMLS_CC);
-				igbinary_serialize16(igsd, (uint16_t) serialized_len TSRMLS_CC);
+				if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser16 TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
+
+				if (igbinary_serialize16(igsd, (uint16_t) serialized_len TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
 			} else {
-				igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser32 TSRMLS_CC);
-				igbinary_serialize32(igsd, (uint32_t) serialized_len TSRMLS_CC);
+				if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_object_ser32 TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
+
+				if (igbinary_serialize32(igsd, (uint32_t) serialized_len TSRMLS_CC) != 0) {
+					if (serialized_data) {
+						efree(serialized_data);
+					}
+					return 1;
+				}
 			}
 
 			if (igbinary_serialize_resize(igsd, serialized_len TSRMLS_CC)) {
 				if (serialized_data) {
 					efree(serialized_data);
 				}
-				r = 1;
 
-				return r;
+				return 1;
 			}
 
 			memcpy(igsd->buffer+igsd->buffer_size, serialized_data, serialized_len);
@@ -1156,7 +1448,7 @@ inline static int igbinary_serialize_object(struct igbinary_serialize_data *igsd
 			r = 1;
 		} else {
 			/* Serialization callback failed, assume null output */
-			igbinary_serialize_null(igsd TSRMLS_CC);
+			r = igbinary_serialize_null(igsd TSRMLS_CC);
 		}
 
 		if (serialized_data) {
@@ -1168,7 +1460,10 @@ inline static int igbinary_serialize_object(struct igbinary_serialize_data *igsd
 
 	/* serialize class name */
 	PHP_SET_CLASS_ATTRIBUTES(z);
-	igbinary_serialize_object_name(igsd, class_name, name_len TSRMLS_CC);
+	if (igbinary_serialize_object_name(igsd, class_name, name_len TSRMLS_CC) != 0) {
+		PHP_CLEANUP_CLASS_ATTRIBUTES();
+		return 1;
+	}
 	PHP_CLEANUP_CLASS_ATTRIBUTES();
 
 	if (ce && ce != PHP_IC_ENTRY && zend_hash_exists(&ce->function_table, "__sleep", sizeof("__sleep"))) {
@@ -1191,8 +1486,10 @@ inline static int igbinary_serialize_object(struct igbinary_serialize_data *igsd
 							"serialize");
 
 					/* empty array */
-					igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC);
-					r = igbinary_serialize8(igsd, 0 TSRMLS_CC);
+					r = igbinary_serialize8(igsd, igbinary_type_array8 TSRMLS_CC);
+					if (r == 0) {
+						r = igbinary_serialize8(igsd, 0 TSRMLS_CC);
+					}
 				}
 			}
 		} else {
@@ -1214,7 +1511,25 @@ inline static int igbinary_serialize_object(struct igbinary_serialize_data *igsd
 /** Serialize zval. */
 static int igbinary_serialize_zval(struct igbinary_serialize_data *igsd, zval *z TSRMLS_DC) {
 	if (Z_ISREF_P(z)) {
-		igbinary_serialize8(igsd, (uint8_t) igbinary_type_ref TSRMLS_CC);
+		if (igbinary_serialize8(igsd, (uint8_t) igbinary_type_ref TSRMLS_CC) != 0) {
+			return 1;
+		}
+
+		/* Complex types serialize a reference, scalars do not... */
+		/* FIXME: Absolutely wrong level to check this. */
+		switch (Z_TYPE_P(z)) {
+			case IS_RESOURCE:
+			case IS_STRING:
+			case IS_LONG:
+			case IS_NULL:
+			case IS_BOOL:
+			case IS_DOUBLE:
+				/* Serialize a reference if zval already added */
+				if (igbinary_serialize_array_ref(igsd, z, false TSRMLS_CC) == 0) {
+					return 0;
+				}
+				/* otherwise fall through */
+		}
 	}
 	switch (Z_TYPE_P(z)) {
 		case IS_RESOURCE:
@@ -1720,6 +2035,8 @@ inline static int igbinary_unserialize_array(struct igbinary_unserialize_data *i
 /** Unserializes object's property array of objects implementing Serializable -interface. */
 inline static int igbinary_unserialize_object_ser(struct igbinary_unserialize_data *igsd, enum igbinary_type t, zval **z, zend_class_entry *ce TSRMLS_DC) {
 	size_t n;
+	int ret;
+	php_unserialize_data_t var_hash;
 
 	if (ce->unserialize == NULL) {
 		zend_error(E_WARNING, "Class %s has no unserializer", ce->name);
@@ -1754,9 +2071,13 @@ inline static int igbinary_unserialize_object_ser(struct igbinary_unserialize_da
 		return 1;
 	}
 
-	if (ce->unserialize(z, ce, (const unsigned char*)(igsd->buffer + igsd->buffer_offset), n, NULL TSRMLS_CC) != SUCCESS) {
-		return 1;
-	} else if (EG(exception)) {
+	PHP_VAR_UNSERIALIZE_INIT(var_hash);
+	ret = ce->unserialize(z, ce,
+		(const unsigned char*)(igsd->buffer + igsd->buffer_offset), n,
+		(zend_unserialize_data *)&var_hash TSRMLS_CC);
+	PHP_VAR_UNSERIALIZE_DESTROY(var_hash);
+
+	if (ret != SUCCESS || EG(exception)) {
 		return 1;
 	}
 
@@ -1978,6 +2299,28 @@ static int igbinary_unserialize_zval(struct igbinary_unserialize_data *igsd, zva
 			if (igbinary_unserialize_zval(igsd, z TSRMLS_CC)) {
 				return 1;
 			}
+			/* Scalar types should be added to the references hash */
+			/* unless they're already added */
+			/* in references list: marked as ref */
+			if (!Z_ISREF_PP(z)) switch (Z_TYPE_PP(z)) {
+				case IS_STRING:
+				case IS_LONG:
+				case IS_NULL:
+				case IS_DOUBLE:
+				case IS_BOOL:
+					/* reference */
+					if (igsd->references_count + 1 >= igsd->references_capacity) {
+						while (igsd->references_count + 1 >= igsd->references_capacity) {
+							igsd->references_capacity *= 2;
+						}
+
+						igsd->references = (void **) erealloc(igsd->references, sizeof(void *) * igsd->references_capacity);
+						if (igsd->references == NULL)
+							return 1;
+					}
+
+					igsd->references[igsd->references_count++] = (void *) *z;
+			}
 			Z_SET_ISREF_TO_PP(z, true);
 			break;
 		case igbinary_type_objref8:
